# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Keras depthwise separable 1D convolution."""


import tensorflow.compat.v2 as tf

from keras import activations
from keras import constraints
from keras import initializers
from keras import regularizers
from keras.layers.convolutional.base_separable_conv import SeparableConv
from keras.utils import conv_utils

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export(
    "keras.layers.SeparableConv1D", "keras.layers.SeparableConvolution1D"
)
class SeparableConv1D(SeparableConv):
    """Depthwise separable 1D convolution.

    This layer performs a depthwise convolution that acts separately on
    channels, followed by a pointwise convolution that mixes channels.
    If `use_bias` is True and a bias initializer is provided,
    it adds a bias vector to the output.
    It then optionally applies an activation function to produce the final
    output.

    Args:
      filters: Integer, the dimensionality of the output space (i.e. the number
        of filters in the convolution).
      kernel_size: A single integer specifying the spatial
        dimensions of the filters.
      strides: A single integer specifying the strides
        of the convolution.
        Specifying any `stride` value != 1 is incompatible with specifying
        any `dilation_rate` value != 1.
      padding: One of `"valid"`, `"same"`, or `"causal"` (case-insensitive).
        `"valid"` means no padding. `"same"` results in padding with zeros
        evenly to the left/right or up/down of the input such that output has
        the same height/width dimension as the input. `"causal"` results in
        causal (dilated) convolutions, e.g. `output[t]` does not depend on
        `input[t+1:]`.
      data_format: A string, one of `channels_last` (default) or
        `channels_first`.  The ordering of the dimensions in the inputs.
        `channels_last` corresponds to inputs with shape
        `(batch_size, length, channels)` while `channels_first` corresponds to
        inputs with shape `(batch_size, channels, length)`.
      dilation_rate: A single integer, specifying
        the dilation rate to use for dilated convolution.
      depth_multiplier: The number of depthwise convolution output channels for
        each input channel. The total number of depthwise convolution output
        channels will be equal to `num_filters_in * depth_multiplier`.
      activation: Activation function to use.
        If you don't specify anything, no activation is applied
        (see `keras.activations`).
      use_bias: Boolean, whether the layer uses a bias.
      depthwise_initializer: An initializer for the depthwise convolution kernel
        (see `keras.initializers`). If None, then the default initializer
        ('glorot_uniform') will be used.
      pointwise_initializer: An initializer for the pointwise convolution kernel
        (see `keras.initializers`). If None, then the default initializer
        ('glorot_uniform') will be used.
      bias_initializer: An initializer for the bias vector. If None, the default
        initializer ('zeros') will be used (see `keras.initializers`).
      depthwise_regularizer: Optional regularizer for the depthwise
        convolution kernel (see `keras.regularizers`).
      pointwise_regularizer: Optional regularizer for the pointwise
        convolution kernel (see `keras.regularizers`).
      bias_regularizer: Optional regularizer for the bias vector
        (see `keras.regularizers`).
      activity_regularizer: Optional regularizer function for the output
        (see `keras.regularizers`).
      depthwise_constraint: Optional projection function to be applied to the
        depthwise kernel after being updated by an `Optimizer` (e.g. used for
        norm constraints or value constraints for layer weights). The function
        must take as input the unprojected variable and must return the
        projected variable (which must have the same shape). Constraints are
        not safe to use when doing asynchronous distributed training
        (see `keras.constraints`).
      pointwise_constraint: Optional projection function to be applied to the
        pointwise kernel after being updated by an `Optimizer`
        (see `keras.constraints`).
      bias_constraint: Optional projection function to be applied to the
        bias after being updated by an `Optimizer`
        (see `keras.constraints`).
      trainable: Boolean, if `True` the weights of this layer will be marked as
        trainable (and listed in `layer.trainable_weights`).

    Input shape:
      3D tensor with shape:
      `(batch_size, channels, steps)` if data_format='channels_first'
      or 3D tensor with shape:
      `(batch_size, steps, channels)` if data_format='channels_last'.

    Output shape:
      3D tensor with shape:
      `(batch_size, filters, new_steps)` if data_format='channels_first'
      or 3D tensor with shape:
      `(batch_size,  new_steps, filters)` if data_format='channels_last'.
      `new_steps` value might have changed due to padding or strides.

    Returns:
      A tensor of rank 3 representing
      `activation(separableconv1d(inputs, kernel) + bias)`.
    """

    def __init__(
        self,
        filters,
        kernel_size,
        strides=1,
        padding="valid",
        data_format=None,
        dilation_rate=1,
        depth_multiplier=1,
        activation=None,
        use_bias=True,
        depthwise_initializer="glorot_uniform",
        pointwise_initializer="glorot_uniform",
        bias_initializer="zeros",
        depthwise_regularizer=None,
        pointwise_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        depthwise_constraint=None,
        pointwise_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            rank=1,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            depth_multiplier=depth_multiplier,
            activation=activations.get(activation),
            use_bias=use_bias,
            depthwise_initializer=initializers.get(depthwise_initializer),
            pointwise_initializer=initializers.get(pointwise_initializer),
            bias_initializer=initializers.get(bias_initializer),
            depthwise_regularizer=regularizers.get(depthwise_regularizer),
            pointwise_regularizer=regularizers.get(pointwise_regularizer),
            bias_regularizer=regularizers.get(bias_regularizer),
            activity_regularizer=regularizers.get(activity_regularizer),
            depthwise_constraint=constraints.get(depthwise_constraint),
            pointwise_constraint=constraints.get(pointwise_constraint),
            bias_constraint=constraints.get(bias_constraint),
            **kwargs
        )

    def call(self, inputs):
        if self.padding == "causal":
            inputs = tf.pad(inputs, self._compute_causal_padding(inputs))
        if self.data_format == "channels_last":
            strides = (1,) + self.strides * 2 + (1,)
            spatial_start_dim = 1
        else:
            strides = (1, 1) + self.strides * 2
            spatial_start_dim = 2

        # Explicitly broadcast inputs and kernels to 4D.
        # TODO(fchollet): refactor when a native separable_conv1d op is
        # available.
        inputs = tf.expand_dims(inputs, spatial_start_dim)
        depthwise_kernel = tf.expand_dims(self.depthwise_kernel, 0)
        pointwise_kernel = tf.expand_dims(self.pointwise_kernel, 0)
        dilation_rate = (1,) + self.dilation_rate

        if self.padding == "causal":
            op_padding = "valid"
        else:
            op_padding = self.padding
        outputs = tf.compat.v1.nn.separable_conv2d(
            inputs,
            depthwise_kernel,
            pointwise_kernel,
            strides=strides,
            padding=op_padding.upper(),
            rate=dilation_rate,
            data_format=conv_utils.convert_data_format(
                self.data_format, ndim=4
            ),
        )

        if self.use_bias:
            outputs = tf.nn.bias_add(
                outputs,
                self.bias,
                data_format=conv_utils.convert_data_format(
                    self.data_format, ndim=4
                ),
            )

        outputs = tf.squeeze(outputs, [spatial_start_dim])

        if self.activation is not None:
            return self.activation(outputs)
        return outputs


# Alias

SeparableConvolution1D = SeparableConv1D



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the RepeatVector layer."""


import tensorflow.compat.v2 as tf

from keras import backend
from keras.engine.base_layer import Layer
from keras.engine.input_spec import InputSpec

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.layers.RepeatVector")
class RepeatVector(Layer):
    """Repeats the input n times.

    Example:

    ```python
    model = Sequential()
    model.add(Dense(32, input_dim=32))
    # now: model.output_shape == (None, 32)
    # note: `None` is the batch dimension

    model.add(RepeatVector(3))
    # now: model.output_shape == (None, 3, 32)
    ```

    Args:
      n: Integer, repetition factor.
    Input shape: 2D tensor of shape `(num_samples, features)`.
    Output shape: 3D tensor of shape `(num_samples, n, features)`.
    """

    def __init__(self, n, **kwargs):
        super().__init__(**kwargs)
        self.n = n
        if not isinstance(n, int):
            raise TypeError(
                f"Expected an integer value for `n`, got {type(n)}."
            )
        self.input_spec = InputSpec(ndim=2)

    def compute_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape).as_list()
        return tf.TensorShape([input_shape[0], self.n, input_shape[1]])

    def call(self, inputs):
        return backend.repeat(inputs, self.n)

    def get_config(self):
        config = {"n": self.n}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains private utilities used mainly by the base Layer class."""

import functools
import threading

import tensorflow.compat.v1 as tf1
import tensorflow.compat.v2 as tf

from keras import backend
from keras.dtensor import dtensor_api as dtensor
from keras.utils import control_flow_util
from keras.utils import tf_inspect
from keras.utils import tf_utils

# isort: off
from tensorflow.python.util.tf_export import keras_export

_call_context = threading.local()


def create_mean_metric(value, name=None):
    # import keras will import base_layer and then this module, and metric
    # relies on base_layer, which result into a cyclic dependency.
    from keras import metrics as metrics_module

    metric_obj = metrics_module.Mean(name=name, dtype=value.dtype)
    return metric_obj, metric_obj(value)


def make_variable(
    name,
    shape=None,
    dtype=tf.float32,
    initializer=None,
    trainable=None,
    caching_device=None,
    validate_shape=True,
    constraint=None,
    use_resource=None,
    collections=None,
    synchronization=tf.VariableSynchronization.AUTO,
    aggregation=tf.VariableAggregation.NONE,
    partitioner=None,
    layout=None,
):
    """Util to create a variable (relies on `variable_scope.variable`).

    Some reuse-related technicalities prevent us from using
    `variable_scope.get_variable()` directly, so we use a subcomponent
    that has fewer constraints (`variable_scope.variable()`).

    In the longer term, it seems like a similar "default variable creator"
    method should exist in `Trackable` instead. When this happens, we can get
    rid of this temporary solution.

    TODO(fchollet): remove this method when no longer needed.

    Args:
      name: Variable name.
      shape: Variable shape.
      dtype: The type of the variable. Defaults to `self.dtype` or `float32`.
      initializer: Initializer instance (callable).
      trainable: Whether the variable should be part of the layer's
        "trainable_variables" (e.g. variables, biases)
        or "non_trainable_variables" (e.g. BatchNorm mean, stddev).
        Note, if the current variable scope is marked as non-trainable
        then this parameter is ignored and any added variables are also
        marked as non-trainable. `trainable` defaults to `True` unless
        `synchronization` is set to `ON_READ`.
      caching_device: Passed to `tf.Variable`.
      validate_shape: Passed to `tf.Variable`.
      constraint: Constraint instance (callable).
      use_resource: Whether to use a `ResourceVariable`.
      collections: List of graph collections keys. The new variable is added to
        these collections. Defaults to `[GraphKeys.GLOBAL_VARIABLES]`.
      synchronization: Indicates when a distributed a variable will be
        aggregated. Accepted values are constants defined in the class
        `tf.VariableSynchronization`. By default the synchronization is set to
        `AUTO` and the current `DistributionStrategy` chooses
        when to synchronize. If `synchronization` is set to `ON_READ`,
        `trainable` must not be set to `True`.
      aggregation: Indicates how a distributed variable will be aggregated.
        Accepted values are constants defined in the class
        `tf.VariableAggregation`.
      partitioner: Not handled at this time.
      layout: the optional DTensor layout, used for creating DVariable.

    Returns:
      Variable instance.
    """
    initializing_from_value = False
    if initializer is not None and not callable(initializer):
        initializing_from_value = True

    if initializing_from_value:
        init_val = initializer
        variable_dtype = None
    else:
        # Instantiate initializer if provided initializer is a type object.
        if tf_inspect.isclass(initializer):
            initializer = initializer()
        if layout:
            init_val = functools.partial(
                initializer, shape, dtype=dtype, layout=layout
            )
        else:
            init_val = functools.partial(initializer, shape, dtype=dtype)
        variable_dtype = dtype.base_dtype

    variable_shape = tf.TensorShape(shape)

    if use_resource is None:
        use_resource = True

    if layout is None:
        # In theory, in `use_resource` is True and `collections` is empty
        # (that is to say, in TF2), we can use tf.Variable.
        # However, this breaks legacy (Estimator) checkpoints because
        # it changes variable names. Remove this when V1 is fully deprecated.
        return tf1.Variable(
            initial_value=init_val,
            name=name,
            trainable=trainable,
            caching_device=caching_device,
            dtype=variable_dtype,
            validate_shape=validate_shape,
            constraint=constraint,
            use_resource=use_resource,
            collections=collections,
            synchronization=synchronization,
            aggregation=aggregation,
            shape=variable_shape if variable_shape else None,
        )
    else:
        return dtensor.DVariable(
            initial_value=init_val,
            name=name,
            trainable=trainable,
            caching_device=caching_device,
            dtype=variable_dtype,
            validate_shape=validate_shape,
            constraint=constraint,
            collections=collections,
            synchronization=synchronization,
            aggregation=aggregation,
            shape=variable_shape if variable_shape else None,
        )


def collect_previous_mask(input_tensors):
    """Retrieves the output mask(s) of the previous node.

    Args:
        input_tensors: An arbitrary structure of Tensors.

    Returns:
        A mask tensor or list of mask tensors.
    """

    def _collect_previous_mask(x):
        return getattr(x, "_keras_mask", None)

    return tf.nest.map_structure(_collect_previous_mask, input_tensors)


def have_all_keras_metadata(tensors):
    return all(hasattr(x, "_keras_history") for x in tf.nest.flatten(tensors))


def generate_placeholders_from_shape(shape):
    return tf1.placeholder(shape=shape, dtype=backend.floatx())


def create_keras_history(tensors):
    """Wraps TensorFlow Operations for compatibility with the Functional API.

    This method checks to see if a Tensor in `tensors` is missing Keras metadata
    and has its origin in a Keras `Input` Layer. If so, this method will replace
    the raw TensorFlow Operations that created this tensor with
    `TensorFlowOpLayer` instances that create identical operations.

    Any Tensors not originating from a Keras `Input` Layer will be treated as
    constants when constructing `TensorFlowOpLayer` instances.

    Args:
      tensors: A structure of Tensors, some of which come from raw TensorFlow
        operations and need to have Keras metadata assigned to them.

    Returns:
      created_layers: List. The `TensorFlowOpLayer` instances created to wrap
        the raw Tensorflow operations.
    """
    _, created_layers = _create_keras_history_helper(tensors, set(), [])
    return created_layers


# Unsafe Internal attribute.
# If True, Keras will not evaluate the constant-foldable inputs to tf op
# layers in TF1 graphs. This *might* speed up model construction time in
# certain settings, but it means
# the models will not be serializable/deserializable via get_config
# (Only via Savedmodels). It may also change the semantics of whether
# generated random numbers are generated once and re-used, or recomputed
# each time.
# Note: This path triggers for TPUEstimators / xla compiled graphs regardless
# of this setting.
_UNSAFE_GRAPH_OP_LAYER_CREATION = False


def _create_keras_history_helper(tensors, processed_ops, created_layers):
    """Helper method for `create_keras_history`.

    Args:
      tensors: A structure of Tensors for which to create Keras metadata.
      processed_ops: Set. TensorFlow operations that have already been wrapped
        in `TensorFlowOpLayer` instances.
      created_layers: List. The `TensorFlowOpLayer` instances created.

    Returns:
      Tuple. First element is the updated set of TensorFlow Operations that
      have been wrapped in `TensorFlowOpLayer` instances. Second element is
      a list of the `TensorFlowOpLayer` instances created.
    """
    if tf1.executing_eagerly_outside_functions():
        raise ValueError(
            "`create_keras_history` should only be called if eager is disabled!"
        )
    # Import of `base_layer` needed in order to create `TensorFlowOpLayer`.
    # Cannot be imported at top because of circular dependencies.
    # TODO(omalleyt): Resolve circular dependency.
    from keras.engine import base_layer

    tensor_list = tf.nest.flatten(tensors)
    sparse_ops = []
    ragged_tensors = []
    for tensor in tensor_list:
        if getattr(tensor, "_keras_history", None) is not None:
            continue
        if isinstance(tensor, (tf.SparseTensor, tf1.SparseTensorValue)):
            sparse_ops.append(tensor.op)
            continue
        if tf_utils.is_ragged(tensor):
            # Ragged tensors don't have an op property
            ragged_tensors.append(tensor)
            continue
        op = tensor.op  # The Op that created this Tensor.
        if op not in processed_ops:
            # Recursively set `_keras_history`.
            op_inputs = list(op.inputs)
            constants = {}
            layer_inputs = []
            for i, op_input in enumerate(op_inputs):
                if uses_keras_history(op_input):
                    layer_inputs.append(op_input)
                else:
                    # Treat any value not originating from a `keras.Input` as
                    # a constant. Variables cannot be supported.
                    ds_with_session = (
                        tf.distribute.in_cross_replica_context()
                        and not tf1.executing_eagerly_outside_functions()
                    )
                    using_xla = control_flow_util.GraphOrParentsInXlaContext(
                        tf1.get_default_graph()
                    )
                    if (
                        ds_with_session
                        or using_xla
                        or _UNSAFE_GRAPH_OP_LAYER_CREATION
                    ):
                        # In Legacy Graph mode, evaluating here makes Session be
                        # configured improperly. The downside of this is that
                        # saving via `get_config` breaks, but SavedModel still
                        # works.
                        constants[i] = op_input
                    else:
                        with tf.init_scope():
                            constants[i] = backend.function([], op_input)([])
            layer_inputs = unnest_if_single_tensor(layer_inputs)
            processed_ops, created_layers = _create_keras_history_helper(
                layer_inputs, processed_ops, created_layers
            )
            name = op.name
            node_def = op.node_def.SerializeToString()
            op_layer = base_layer.TensorFlowOpLayer(
                node_def, constants=constants, name=name
            )
            created_layers.append(op_layer)
            op_layer._set_connectivity_metadata(
                args=(layer_inputs,), kwargs={}, outputs=op.outputs
            )
            processed_ops.update([op])
    if sparse_ops or ragged_tensors:
        lambda_example = """
    weights_mult = lambda x: tf.sparse.sparse_dense_matmul(x, weights)
    output = tf.keras.layers.Lambda(weights_mult)(input)
    """
        raise ValueError(
            "Tensorflow ops that generate ragged or sparse tensor "
            "outputs are currently not supported by Keras automatic "
            "op wrapping. Please wrap these ops in a Lambda layer: "
            "\n\n```\n{example}\n```\n"
            "Sparse ops encountered: {sparse_ops}\n"
            "Ragged tensors encountered: {ragged_tensors}\n".format(
                example=lambda_example,
                sparse_ops=str(sparse_ops),
                ragged_tensors=str(ragged_tensors),
            )
        )
    return processed_ops, created_layers


def unnest_if_single_tensor(input_tensors):
    # Preserve compatibility with older configs
    flat_input_tensors = tf.nest.flatten(input_tensors)
    # If this is a single element but not a dict, unwrap. If this is a dict,
    # assume the first layer expects a dict (as is the case with a
    # DenseFeatures layer); pass through.
    if not isinstance(input_tensors, dict) and len(flat_input_tensors) == 1:
        input_tensors = flat_input_tensors[0]
    return input_tensors


def needs_keras_history(tensors, ignore_call_context=False):
    """Check if any Tensors need to be wrapped in TensorFlowOpLayers.

    This will never return True inside a sublayer, because sublayers
    do not need to create Keras History. Otherwise, this returns True
    if one or more of `tensors` originates from a `keras.Input` and
    does not have `_keras_history` set.

    Args:
      tensors: An arbitrary nested structure of Tensors.
      ignore_call_context: Whether to ignore the check of if currently
        outside of a `call` context. This is `True` when creating
        KerasHistory inside `Node`, where we always know that Tensors
        are being used with the Functional API.

    Returns:
      Bool, whether at least one Tensor needs to be wrapped.
    """
    input_tensors = tf.nest.flatten(tensors)
    if call_context().in_call and not ignore_call_context:
        return False
    if all(
        getattr(tensor, "_keras_history", None) is not None
        for tensor in input_tensors
    ):
        # KerasHistory already set.
        return False
    return uses_keras_history(tensors)


def is_in_keras_graph():
    """Returns if currently executing inside of a Keras graph."""
    return call_context().in_keras_graph


def is_in_eager_or_tf_function():
    """Returns if in eager mode or inside of a tf.function."""
    return tf.executing_eagerly() or is_in_tf_function()


def is_in_tf_function():
    """Returns if inside of a tf.function."""
    # Check if running in V1 graph mode.
    if not tf1.executing_eagerly_outside_functions():
        return False
    if not tf.inside_function():
        return False
    # Check if inside Keras FuncGraph.
    if is_in_keras_graph():
        return False
    # Check for a v1 `wrap_function` FuncGraph.
    graph = tf1.get_default_graph()
    if getattr(graph, "name", False) and graph.name.startswith(
        "wrapped_function"
    ):
        return False
    return True


def uses_keras_history(tensors):
    """Check if at least one Tensor originates from a `keras.Input`.

    This is `True` if at least one Tensor has its origin in a `keras.Input`.
    Any Tensor that originates from a `keras.Input` will have a dependency
    Tensor with a `_keras_history` attribute attached. Tensors that have
    already been checked to not originate from a `keras.Input`
    are marked as `_keras_history_checked`.

    Args:
      tensors: An arbitrary nested structure of Tensors.

    Returns:
      Bool, whether at least one Tensor originates from a `keras.Input`.
    """
    checked_tensors = set()
    tensors_to_check = tf.nest.flatten(tensors)

    while tensors_to_check:
        new_tensors_to_check = []
        for tensor in tensors_to_check:
            if id(tensor) in checked_tensors:
                continue

            checked_tensors.add(id(tensor))

            if getattr(tensor, "_keras_history_checked", None) is not None:
                continue
            if getattr(tensor, "_keras_history", None) is not None:
                return True

            try:
                new_tensors_to_check.extend(tensor.op.inputs)
            except AttributeError:
                # In case `tensor` is a Variable created in an Eager context.
                pass

        tensors_to_check = new_tensors_to_check

    # Mark that these Tensors have been checked once for `_keras_history`,
    # and should not be checked again for performance reasons.
    mark_checked(tensors)
    return False


def mark_checked(tensors):
    """Marks that these Tensors should not be tracked.

    This prevents Layers from attempting to create TensorFlowOpLayers
    for these Tensors.

    Args:
      tensors: An arbitrary structure of Tensors.
    """

    def _mark_checked(tensor):
        tensor._keras_history_checked = True

    tf.nest.map_structure(_mark_checked, tensors)


def call_context():
    """Returns currently active `CallContext`."""
    call_ctx = getattr(_call_context, "call_context", None)
    if call_ctx is None:
        call_ctx = CallContext()
        _call_context.call_context = call_ctx
    return call_ctx


# Inject the call_context function to keras_deps to remove the dependency
# from TFLite to Keras.
tf.__internal__.register_call_context_function(call_context)


class CallContext:
    """Keeps track of properties currently inside a Layer/Model's `call`.

    Attributes:
      in_call: Whether currently inside the `call` of a Layer.
      layer: The `Layer` whose `call` is currently active.
      inputs: The inputs to the currently active `Layer`.
      build_graph: Whether currently inside a Graph or FuncGraph.
      training: Whether currently executing in training or inference mode.
      saving: Whether currently saving to SavedModel.
      frozen: Whether currently executing inside a `Layer` with `trainable` set
        to `False`.
      in_keras_graph: Whether executing inside the Keras Graph.
    """

    def __init__(self):
        # Handle `in_call` separately as it is the most-read attr and reading it
        # is on the hot path.
        self.in_call = False
        self._state = {
            "layer": None,
            "inputs": None,
            "build_graph": False,
            "training": None,
            "saving": None,
        }
        # TODO(b/150169018): This logic can be replaced after the Functional API
        # refactor.
        self._in_keras_graph = False

    def enter(self, layer, inputs, build_graph, training, saving=None):
        """Push a Layer and its inputs and state onto the current call context.

        Args:
          layer: The `Layer` whose `call` is currently active.
          inputs: The inputs to the currently active `Layer`.
          build_graph: Whether currently inside a Graph or FuncGraph.
          training: Whether currently executing in training or inference mode.
          saving: Whether currently saving to SavedModel.

        Returns:
          Context manager.
        """
        state = {
            "layer": layer,
            "inputs": inputs,
            "build_graph": build_graph,
            "training": training,
            "saving": saving,
        }
        return CallContextManager(self, state)

    @property
    def layer(self):
        return self._state["layer"]

    @property
    def inputs(self):
        return self._state["inputs"]

    @property
    def build_graph(self):
        return self._state["build_graph"]

    @property
    def training(self):
        return self._state["training"]

    @property
    def saving(self):
        return self._state["saving"]

    @property
    def frozen(self):
        layer = self._state["layer"]
        if not layer:
            return False
        return not layer.trainable

    @property
    def in_keras_graph(self):
        # Returns True even if in a subgraph of the Keras graph, such as those
        # created by control flow ops.
        if tf.executing_eagerly():
            return False
        return (
            self._in_keras_graph
            or getattr(backend.get_graph(), "name", None) == "keras_graph"
        )


class CallContextManager:
    """Context manager for `CallContext`."""

    def __init__(self, call_ctx, state):
        self._call_ctx = call_ctx
        self._state = state
        self._build_graph = state["build_graph"]

    def __enter__(self):
        call_ctx = self._call_ctx
        self._prev_in_call = call_ctx.in_call
        self._prev_state = call_ctx._state

        call_ctx.in_call = True
        call_ctx._state = self._state

        # TODO(b/150169018): This logic can be removed after the Functional API
        # refactor.
        if self._build_graph:
            self._prev_in_keras_graph = call_ctx._in_keras_graph
            call_ctx._in_keras_graph = (
                call_ctx._in_keras_graph
                or getattr(backend.get_graph(), "name", None) == "keras_graph"
            )

    def __exit__(self, *exc_info):
        call_ctx = self._call_ctx
        call_ctx.in_call = self._prev_in_call
        call_ctx._state = self._prev_state

        if self._build_graph:
            call_ctx._in_keras_graph = self._prev_in_keras_graph


def training_arg_passed_to_call(argspec, args, kwargs):
    """Returns whether a user passed the `training` argument in `__call__`."""
    # `argspec.args` starts with ['self', 'inputs']
    full_args = dict(zip(argspec.args[2:], args))
    full_args.update(kwargs)
    return "training" in full_args and full_args["training"] is not None


def is_subclassed(layer):
    """Returns True if the object is a subclassed layer or subclassed model."""
    return (
        layer.__module__.find("keras.engine") == -1
        and layer.__module__.find("keras.layers") == -1
    )


def from_saved_model(layer):
    """Returns whether the layer is loaded from a SavedModel."""
    return layer.__module__.find("keras.saving.legacy.saved_model") != -1


def check_graph_consistency(tensor=None, method="add_loss", force_raise=False):
    """Checks that tensors passed to `add_*` method match the Keras graph.

    When one of the `add_*` method is called inside a V2 conditional branch, the
    underlying tensor gets created in a FuncGraph managed by control_flow_v2.
    We need to raise clear error messages in such cases.

    Args:
      tensor: Tensor to check, or `False` if it is known that an error
        should be raised.
      method: Caller method, one of {'add_metric', 'add_loss', 'add_update'}.
      force_raise: If an error should be raised regardless of `tensor`.

    Raises:
      RuntimeError: In case of an out-of-graph tensor.
    """
    if force_raise or (
        tf1.executing_eagerly_outside_functions()
        and hasattr(tensor, "graph")
        and tensor.graph.is_control_flow_graph
    ):
        if method == "activity_regularizer":
            bad_example = """
      class TestModel(tf.keras.Model):

        def __init__(self):
          super(TestModel, self).__init__(name='test_model')
          self.dense = tf.keras.layers.Dense(2, activity_regularizer='l2')

        def call(self, x, training=None):
          if training:
            return self.dense(x)
          else:
            return self.dense(x)
      """
            correct_example = """
      class TestModel(tf.keras.Model):

        def __init__(self):
          super(TestModel, self).__init__(name='test_model')
          self.dense = tf.keras.layers.Dense(2, activity_regularizer='l2')

        def call(self, x, training=None):
          return self.dense(x)
      """
            raise RuntimeError(
                "You are using a layer with `activity_regularizer` in a "
                f"control flow branch, e.g.:\n{bad_example}\nThis is currently "
                "not supported. Please move your call to the layer with "
                "`activity_regularizer` out of the control flow branch, "
                f"e.g.:\n{correct_example}\nYou can also resolve this by "
                "marking your outer model/layer dynamic (eager-only) by "
                "passing `dynamic=True` to the layer constructor. Any kind of "
                "control flow is supported with dynamic layers. Note that "
                "using `dynamic=True` requires you to implement static shape "
                "inference in the `compute_output_shape(input_shape)` "
                "method."
            )

        if method == "add_metric":
            bad_example = """
      def call(self, inputs, training=None):
        if training:
          metric = compute_metric(inputs)
          self.add_metric(metric, name='my_metric', aggregation='mean')
        return inputs
      """
            correct_example = """
      def call(self, inputs, training=None):
        if training:
          metric = compute_metric(inputs)
        else:
          metric = 0.
        self.add_metric(metric, name='my_metric', aggregation='mean')
        return inputs
      """
        elif method == "add_loss":
            bad_example = """
      def call(self, inputs, training=None):
        if training:
          loss = compute_loss(inputs)
          self.add_loss(loss)
        return inputs
      """
            correct_example = """
      def call(self, inputs, training=None):
        if training:
          loss = compute_loss(inputs)
        else:
          loss = 0.
        self.add_loss(loss)
        return inputs
      """
        else:
            bad_example = """
      def call(self, inputs, training=None):
        if training:
          self.add_update(self.w.assign_add(1))
        return inputs
      """
            correct_example = """
      def call(self, inputs, training=None):
        if training:
          increment = 1
        else:
          increment = 0
        self.add_update(self.w.assign_add(increment))
        return inputs
      """
        raise RuntimeError(
            "You are using the method `{method}` in a control flow branch "
            "in your layer, e.g.:\n{bad_example}\n"
            "This is not currently supported. "
            "Please move your call to {method} out of the control flow branch, "
            "e.g.:\n{correct_example}\n"
            "You can also resolve this by marking your layer "
            "as dynamic (eager-only) by passing "
            "`dynamic=True` to the layer constructor. "
            "Any kind of control flow is supported with dynamic layers. "
            "Note that using `dynamic=True` requires you "
            "to implement static shape inference "
            "in the `compute_output_shape(input_shape)` method.".format(
                method=method,
                bad_example=bad_example,
                correct_example=correct_example,
            )
        )


def mark_as_return(outputs, acd):
    """Marks `outputs` as the return values for automatic control deps."""

    def _mark_as_return(tensor):
        """Marks `tensor` as the return value for automatic control deps."""
        if not tf.is_tensor(tensor):
            return tensor

        return_tensor = acd.mark_as_return(tensor)
        if getattr(tensor, "_keras_mask", None) is not None:
            return_tensor._keras_mask = acd.mark_as_return(tensor._keras_mask)
        else:
            return_tensor._keras_mask = None

        # Handle TensorFlow Probability attached metadata.
        # TODO(b/132076537): Remove this once TFP uses `CompositeTensor`.
        if getattr(tensor, "_tfp_distribution", None) is not None:
            return_tensor._tfp_distribution = tensor._tfp_distribution

        return return_tensor

    return tf.nest.map_structure(_mark_as_return, outputs)


V2_DTYPE_BEHAVIOR = None


@keras_export(v1=["keras.layers.enable_v2_dtype_behavior"])
def enable_v2_dtype_behavior():
    """Enable the V2 dtype behavior for Keras layers.

    By default, the V2 dtype behavior is enabled in TensorFlow 2, so this
    function is only useful if `tf.compat.v1.disable_v2_behavior` has been
    called. Since mixed precision requires V2 dtype behavior to be enabled, this
    function allows you to use mixed precision in Keras layers if
    `disable_v2_behavior` has been called.

    When enabled, the dtype of Keras layers defaults to floatx (which is
    typically float32) instead of None. In addition, layers will automatically
    cast floating-point inputs to the layer's dtype.

    >>> x = tf.ones((4, 4, 4, 4), dtype='float64')
    >>> layer = tf.keras.layers.Conv2D(filters=4, kernel_size=2)
    >>> print(layer.dtype)  # float32 since V2 dtype behavior is enabled
    float32
    >>> y = layer(x)  # Layer casts inputs since V2 dtype behavior is enabled
    >>> print(y.dtype.name)
    float32

    A layer author can opt-out their layer from the automatic input casting by
    passing `autocast=False` to the base Layer's constructor. This disables the
    autocasting part of the V2 behavior for that layer, but not the defaulting
    to floatx part of the V2 behavior.

    When a global `tf.keras.mixed_precision.Policy` is set, a Keras layer's
    dtype will default to the global policy instead of floatx. Layers will
    automatically cast inputs to the policy's compute_dtype.
    """
    global V2_DTYPE_BEHAVIOR
    V2_DTYPE_BEHAVIOR = True


@keras_export(v1=["keras.layers.disable_v2_dtype_behavior"])
def disable_v2_dtype_behavior():
    """Disables the V2 dtype behavior for Keras layers.

    See `tf.compat.v1.keras.layers.enable_v2_dtype_behavior`.
    """
    global V2_DTYPE_BEHAVIOR
    V2_DTYPE_BEHAVIOR = False


def v2_dtype_behavior_enabled():
    """Returns True if the V2 dtype behavior is enabled."""
    if V2_DTYPE_BEHAVIOR is None:
        return tf.__internal__.tf2.enabled()
    return V2_DTYPE_BEHAVIOR


class TrackableWeightHandler:
    """Keras wrapper for handling Trackable object saving and restoring.

    This class handles Trackables in both V1 and V2 modes, ensuring that they
    can be saved and restored with the correct data and without adding
    additional ops on every save.

    Attributes:
      trackable: The trackable to wrap.
      num_tensors: The number of tensors that this trackable requires for
        saving.
    """

    def __init__(self, trackable):
        if not isinstance(trackable, tf.__internal__.tracking.Trackable):
            raise ValueError(f"{trackable} is not a Trackable object.")
        self._trackable = trackable
        self._distribute_strategy = tf.distribute.get_strategy()

        saveables = tf.__internal__.tracking.saveable_objects_from_trackable(
            trackable
        ).values()
        # 'Saveables' won't exist when we're passed a legacy TF1 table like
        # a StaticHashTable.
        if not saveables:
            self._num_tensors = 0
            self._setter = lambda weights: None
            self._getter = lambda: []

        elif len(saveables) == 1:
            saveable = list(saveables)[0]

            if tf1.executing_eagerly_outside_functions():
                # If we're in eager mode, we need to defer calling the
                # Trackable's saveable() callable until data export time.
                # However, it is safe to call the saveable as many times as we
                # want, so we will call it now to figure out how many tensors
                # this Trackable will produce.
                self._saveable = saveable
                self._num_tensors = len(self._saveable().specs)
                self._setter = lambda weights: self._saveable().restore(
                    weights, None
                )
                self._getter = lambda: [
                    spec.tensor for spec in self._saveable().specs
                ]
            else:
                # If we're in Graph mode, we need to evaluate the Saveable only
                # once and cache the resulting restore graph. Failing to do this
                # will result in new assignment ops being added to the graph
                # each time set_weights() is called.
                self._placeholder_tensors = []
                self._saveable = saveable()
                self._num_tensors = len(self._saveable.specs)
                for spec in self._saveable.specs:
                    tensor = spec.tensor
                    self._placeholder_tensors.append(
                        tf1.placeholder(tensor.dtype, tensor.shape)
                    )
                self._assign_op = self._saveable.restore(
                    self._placeholder_tensors, None
                )
                self._setter = self._set_weights_v1
                self._getter = lambda: [
                    spec.tensor for spec in self._saveable.specs
                ]
        else:
            raise ValueError(
                "Only Trackables with one Saveable are supported. "
                f"The Trackable {trackable} has {len(saveables)} Saveables."
            )

    @property
    def num_tensors(self):
        return self._num_tensors

    def set_weights(self, weights):
        if len(weights) != self._num_tensors:
            raise ValueError(
                f"Weight handler for trackable {self._trackable} received "
                "an incorrect number of weights: "
                f"expected {self._num_tensors} weights, "
                f"got {len(weights)} weights."
            )
        self._setter(weights)

    def get_tensors(self):
        return self._getter()

    def _set_weights_v1(self, weights):
        feed_dict = {}
        for idx, tensor in enumerate(weights):
            feed_dict[self._placeholder_tensors[idx]] = tensor
        backend.get_session().run(self._assign_op, feed_dict)


def no_ragged_support(inputs, layer_name):
    input_list = tf.nest.flatten(inputs)
    if any(isinstance(x, tf.RaggedTensor) for x in input_list):
        raise ValueError(
            f"Layer {layer_name} does not support RaggedTensors as input. "
            f"Inputs received: {inputs}. You can try converting your "
            "input to a dense (uniform) tensor."
        )


def is_split_variable(v):
    """Returns True if `v` is a PartitionedVariable or a ShardedVariable."""
    return not {clz.__name__ for clz in v.__class__.__mro__}.isdisjoint(
        {"PartitionedVariable", "ShardedVariable"}
    )


def has_weights(obj):
    obj_type = type(obj)
    return (
        hasattr(obj_type, "trainable_weights")
        and hasattr(obj_type, "non_trainable_weights")
        and not isinstance(obj, type)
    )


# TODO(kathywu): This is a temporary hack. When a network of layers is revived
# from SavedModel, only the top-level layer will have losses. This causes issues
# in eager mode because the child layers may have graph losses
# (thus model.losses returns a mix of Eager and graph tensors). To fix this,
# whenever eager losses are added to one layer, add eager losses to all
# child layers. This causes `.losses` to only return eager losses.
REVIVED_LOSS_PLACEHOLDER = (
    "This layer's losses have been added to the parent layer."
)



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Correctness tests for tf.keras RNN models using DistributionStrategy."""

import numpy as np
import tensorflow.compat.v2 as tf

import keras
from keras.distribute import keras_correctness_test_base
from keras.layers.rnn import gru
from keras.layers.rnn import gru_v1
from keras.layers.rnn import lstm
from keras.layers.rnn import lstm_v1
from keras.mixed_precision import policy
from keras.optimizers.legacy import gradient_descent as gradient_descent_keras
from keras.testing_infra import test_utils


class _DistributionStrategyRnnModelCorrectnessTest(
    keras_correctness_test_base.TestDistributionStrategyEmbeddingModelCorrectnessBase  # noqa: E501
):
    def _get_layer_class(self):
        raise NotImplementedError

    def get_model(
        self,
        max_words=10,
        initial_weights=None,
        distribution=None,
        input_shapes=None,
    ):
        del input_shapes
        rnn_cls = self._get_layer_class()

        with keras_correctness_test_base.MaybeDistributionScope(distribution):
            word_ids = keras.layers.Input(
                shape=(max_words,), dtype=np.int32, name="words"
            )
            word_embed = keras.layers.Embedding(input_dim=20, output_dim=10)(
                word_ids
            )
            rnn_embed = rnn_cls(units=4, return_sequences=False)(word_embed)

            dense_output = keras.layers.Dense(2)(rnn_embed)
            preds = keras.layers.Softmax(dtype="float32")(dense_output)
            model = keras.Model(inputs=[word_ids], outputs=[preds])

            if initial_weights:
                model.set_weights(initial_weights)

            optimizer_fn = gradient_descent_keras.SGD

            model.compile(
                optimizer=optimizer_fn(learning_rate=0.1),
                loss="sparse_categorical_crossentropy",
                metrics=["sparse_categorical_accuracy"],
            )
        return model


@test_utils.run_all_without_tensor_float_32(
    "Uses Dense layers, which call matmul"
)
class DistributionStrategyGruModelCorrectnessTest(
    _DistributionStrategyRnnModelCorrectnessTest
):
    def _get_layer_class(self):
        if tf.__internal__.tf2.enabled():
            if not tf.executing_eagerly():
                self.skipTest(
                    "GRU v2 and legacy graph mode don't work together."
                )
            return gru.GRU
        else:
            return gru_v1.GRU

    @tf.__internal__.distribute.combinations.generate(
        keras_correctness_test_base.test_combinations_for_embedding_model()
        + keras_correctness_test_base.multi_worker_mirrored_eager()
    )
    def test_gru_model_correctness(
        self, distribution, use_numpy, use_validation_data
    ):
        self.run_correctness_test(distribution, use_numpy, use_validation_data)


@test_utils.run_all_without_tensor_float_32(
    "Uses Dense layers, which call matmul"
)
class DistributionStrategyLstmModelCorrectnessTest(
    _DistributionStrategyRnnModelCorrectnessTest
):
    def _get_layer_class(self):
        if tf.__internal__.tf2.enabled():
            if not tf.executing_eagerly():
                self.skipTest(
                    "LSTM v2 and legacy graph mode don't work together."
                )
            return lstm.LSTM
        else:
            return lstm_v1.LSTM

    @tf.__internal__.distribute.combinations.generate(
        keras_correctness_test_base.test_combinations_for_embedding_model()
        + keras_correctness_test_base.multi_worker_mirrored_eager()
    )
    def test_lstm_model_correctness(
        self, distribution, use_numpy, use_validation_data
    ):
        self.run_correctness_test(distribution, use_numpy, use_validation_data)

    @tf.__internal__.distribute.combinations.generate(
        keras_correctness_test_base.test_combinations_for_embedding_model()
        + keras_correctness_test_base.multi_worker_mirrored_eager()
    )
    @test_utils.enable_v2_dtype_behavior
    def test_lstm_model_correctness_mixed_precision(
        self, distribution, use_numpy, use_validation_data
    ):
        if isinstance(
            distribution,
            (
                tf.distribute.experimental.CentralStorageStrategy,
                tf.compat.v1.distribute.experimental.CentralStorageStrategy,
            ),
        ):
            self.skipTest(
                "CentralStorageStrategy is not supported by mixed precision."
            )
        if isinstance(
            distribution,
            (
                tf.distribute.experimental.TPUStrategy,
                tf.compat.v1.distribute.experimental.TPUStrategy,
            ),
        ):
            policy_name = "mixed_bfloat16"
        else:
            policy_name = "mixed_float16"

        with policy.policy_scope(policy_name):
            self.run_correctness_test(
                distribution, use_numpy, use_validation_data
            )


if __name__ == "__main__":
    tf.__internal__.distribute.multi_process_runner.test_main()



"""Model that incorporates a set of edge case development patterns.
"""

import tensorflow as tf
from tensorflow import keras

from keras.integration_test.models.input_spec import InputSpec

INPUT_DIM = 32
NUM_CLASSES = 5


def get_data_spec(batch_size):
    return (
        InputSpec((batch_size, INPUT_DIM)),
        InputSpec((batch_size, NUM_CLASSES)),
    )


def get_input_preprocessor():
    return None


class LinearA(keras.layers.Layer):
    """Standard custom layer with 2 call() inputs."""

    def __init__(self, units=32, input_dim=32):
        super().__init__()
        self.w = self.add_weight(
            shape=(input_dim, units),
            initializer="random_normal",
            trainable=True,
        )
        self.b = self.add_weight(
            shape=(units,), initializer="zeros", trainable=True
        )

    def call(self, inputs_1, inputs_2):
        return (
            tf.matmul(inputs_1, self.w) + tf.matmul(inputs_2, self.w) + self.b
        )


class LinearB(keras.layers.Layer):
    """Layer that tracks weights in a dict attribute that gets updated later."""

    def __init__(self, units=32, input_dim=32, **kwargs):
        super().__init__(**kwargs)
        w_init = tf.random_normal_initializer()
        b_init = tf.zeros_initializer()
        self.state = {
            "kernel": tf.Variable(
                initial_value=w_init(shape=(input_dim, units), dtype="float32"),
                trainable=True,
                name="kernel",
            )
        }
        self.state["bias"] = tf.Variable(
            initial_value=b_init(shape=(units,), dtype="float32"),
            trainable=True,
            name="bias",
        )

    def call(self, inputs):
        return tf.matmul(inputs, self.state["kernel"]) + self.state["bias"]


class LinearC(keras.layers.Layer):
    """Layer that creates weights in call()."""

    def __init__(self, units=32, input_dim=32, **kwargs):
        super().__init__(**kwargs)
        self._custom_built = False
        self.units = units
        self.input_dim = input_dim

    def call(self, inputs):
        if not self._custom_built:
            self.w = self.add_weight(
                shape=(self.input_dim, self.units),
                initializer="random_normal",
                trainable=True,
            )
            self.b = self.add_weight(
                shape=(self.units,), initializer="zeros", trainable=True
            )
            self._custom_built = True
        return tf.matmul(inputs, self.w) + self.b


class BatchNorm(keras.layers.Layer):
    """Layer with different training/test behavior and non-trainable updates."""

    def __init__(
        self, scale=True, center=True, epsilon=1e-6, momentum=0.9, **kwargs
    ):
        super().__init__(**kwargs)
        self.scale = scale
        self.center = center
        self.epsilon = epsilon
        self.momentum = momentum

    def build(self, input_shape):
        self.var = self.add_weight(
            shape=[input_shape[1]], initializer="ones", trainable=False
        )
        self.mean = self.add_weight(
            shape=[input_shape[1]], initializer="zeros", trainable=False
        )
        self.gamma = self.add_weight(shape=[input_shape[1]], initializer="ones")
        self.beta = self.add_weight(shape=[input_shape[1]], initializer="zeros")

    def call(self, inputs, training=False):
        if training:
            mean, var = tf.nn.moments(inputs, axes=[0])
            outputs = (inputs - mean) / (var + self.epsilon)
            self.var.assign(self.var * self.momentum + var * 0.1)
            self.mean.assign(self.mean * self.momentum + mean * 0.1)
        else:
            outputs = (inputs - self.mean) / (self.var + self.epsilon)
        if self.scale:
            outputs *= self.gamma
        if self.center:
            outputs += self.beta
        return outputs


class FunctionalSubclassModel(keras.Model):
    def __init__(self, **kwargs):
        inputs = keras.Input((INPUT_DIM,))
        x = inputs
        x = LinearA(32, INPUT_DIM)(x, x)
        x = LinearB(32, 32)(x)
        x = LinearC(32, 32)(x)
        x = BatchNorm()(x)
        outputs = keras.layers.Dense(NUM_CLASSES, activation="softmax")(x)
        super().__init__(inputs, outputs, **kwargs)


def get_model(
    build=False, compile=False, jit_compile=False, include_preprocessing=True
):
    model = FunctionalSubclassModel()
    if compile:
        model.compile("rmsprop", "mse", jit_compile=jit_compile)
    return model


def get_custom_objects():
    return {
        "LinearA": LinearA,
        "LinearB": LinearB,
        "LinearC": LinearC,
        "BatchNorm": BatchNorm,
    }



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""ResNet v2 models for Keras.

Reference:
  - [Identity Mappings in Deep Residual Networks](
      https://arxiv.org/abs/1603.05027) (CVPR 2016)
"""

from keras.applications import imagenet_utils
from keras.applications import resnet

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export(
    "keras.applications.resnet_v2.ResNet50V2", "keras.applications.ResNet50V2"
)
def ResNet50V2(
    include_top=True,
    weights="imagenet",
    input_tensor=None,
    input_shape=None,
    pooling=None,
    classes=1000,
    classifier_activation="softmax",
):
    """Instantiates the ResNet50V2 architecture."""

    def stack_fn(x):
        x = resnet.stack2(x, 64, 3, name="conv2")
        x = resnet.stack2(x, 128, 4, name="conv3")
        x = resnet.stack2(x, 256, 6, name="conv4")
        return resnet.stack2(x, 512, 3, stride1=1, name="conv5")

    return resnet.ResNet(
        stack_fn,
        True,
        True,
        "resnet50v2",
        include_top,
        weights,
        input_tensor,
        input_shape,
        pooling,
        classes,
        classifier_activation=classifier_activation,
    )


@keras_export(
    "keras.applications.resnet_v2.ResNet101V2", "keras.applications.ResNet101V2"
)
def ResNet101V2(
    include_top=True,
    weights="imagenet",
    input_tensor=None,
    input_shape=None,
    pooling=None,
    classes=1000,
    classifier_activation="softmax",
):
    """Instantiates the ResNet101V2 architecture."""

    def stack_fn(x):
        x = resnet.stack2(x, 64, 3, name="conv2")
        x = resnet.stack2(x, 128, 4, name="conv3")
        x = resnet.stack2(x, 256, 23, name="conv4")
        return resnet.stack2(x, 512, 3, stride1=1, name="conv5")

    return resnet.ResNet(
        stack_fn,
        True,
        True,
        "resnet101v2",
        include_top,
        weights,
        input_tensor,
        input_shape,
        pooling,
        classes,
        classifier_activation=classifier_activation,
    )


@keras_export(
    "keras.applications.resnet_v2.ResNet152V2", "keras.applications.ResNet152V2"
)
def ResNet152V2(
    include_top=True,
    weights="imagenet",
    input_tensor=None,
    input_shape=None,
    pooling=None,
    classes=1000,
    classifier_activation="softmax",
):
    """Instantiates the ResNet152V2 architecture."""

    def stack_fn(x):
        x = resnet.stack2(x, 64, 3, name="conv2")
        x = resnet.stack2(x, 128, 8, name="conv3")
        x = resnet.stack2(x, 256, 36, name="conv4")
        return resnet.stack2(x, 512, 3, stride1=1, name="conv5")

    return resnet.ResNet(
        stack_fn,
        True,
        True,
        "resnet152v2",
        include_top,
        weights,
        input_tensor,
        input_shape,
        pooling,
        classes,
        classifier_activation=classifier_activation,
    )


@keras_export("keras.applications.resnet_v2.preprocess_input")
def preprocess_input(x, data_format=None):
    return imagenet_utils.preprocess_input(
        x, data_format=data_format, mode="tf"
    )


@keras_export("keras.applications.resnet_v2.decode_predictions")
def decode_predictions(preds, top=5):
    return imagenet_utils.decode_predictions(preds, top=top)


preprocess_input.__doc__ = imagenet_utils.PREPROCESS_INPUT_DOC.format(
    mode="",
    ret=imagenet_utils.PREPROCESS_INPUT_RET_DOC_TF,
    error=imagenet_utils.PREPROCESS_INPUT_ERROR_DOC,
)
decode_predictions.__doc__ = imagenet_utils.decode_predictions.__doc__

DOC = """

  Reference:
  - [Identity Mappings in Deep Residual Networks](
      https://arxiv.org/abs/1603.05027) (CVPR 2016)

  For image classification use cases, see
  [this page for detailed examples](
    https://keras.io/api/applications/#usage-examples-for-image-classification-models).

  For transfer learning use cases, make sure to read the
  [guide to transfer learning & fine-tuning](
    https://keras.io/guides/transfer_learning/).

  Note: each Keras Application expects a specific kind of input preprocessing.
  For ResNetV2, call `tf.keras.applications.resnet_v2.preprocess_input` on your
  inputs before passing them to the model.
  `resnet_v2.preprocess_input` will scale input pixels between -1 and 1.

  Args:
    include_top: whether to include the fully-connected
      layer at the top of the network.
    weights: one of `None` (random initialization),
      'imagenet' (pre-training on ImageNet),
      or the path to the weights file to be loaded.
    input_tensor: optional Keras tensor (i.e. output of `layers.Input()`)
      to use as image input for the model.
    input_shape: optional shape tuple, only to be specified
      if `include_top` is False (otherwise the input shape
      has to be `(224, 224, 3)` (with `'channels_last'` data format)
      or `(3, 224, 224)` (with `'channels_first'` data format).
      It should have exactly 3 inputs channels,
      and width and height should be no smaller than 32.
      E.g. `(200, 200, 3)` would be one valid value.
    pooling: Optional pooling mode for feature extraction
      when `include_top` is `False`.
      - `None` means that the output of the model will be
          the 4D tensor output of the
          last convolutional block.
      - `avg` means that global average pooling
          will be applied to the output of the
          last convolutional block, and thus
          the output of the model will be a 2D tensor.
      - `max` means that global max pooling will
          be applied.
    classes: optional number of classes to classify images
      into, only to be specified if `include_top` is True, and
      if no `weights` argument is specified.
    classifier_activation: A `str` or callable. The activation function to use
      on the "top" layer. Ignored unless `include_top=True`. Set
      `classifier_activation=None` to return the logits of the "top" layer.
      When loading pretrained weights, `classifier_activation` can only
      be `None` or `"softmax"`.

  Returns:
    A `keras.Model` instance.
"""

setattr(ResNet50V2, "__doc__", ResNet50V2.__doc__ + DOC)
setattr(ResNet101V2, "__doc__", ResNet101V2.__doc__ + DOC)
setattr(ResNet152V2, "__doc__", ResNet152V2.__doc__ + DOC)



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Keras upsampling layer for 1D inputs."""


import tensorflow.compat.v2 as tf

from keras import backend
from keras.engine.base_layer import Layer
from keras.engine.input_spec import InputSpec

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.layers.UpSampling1D")
class UpSampling1D(Layer):
    """Upsampling layer for 1D inputs.

    Repeats each temporal step `size` times along the time axis.

    Examples:

    >>> input_shape = (2, 2, 3)
    >>> x = np.arange(np.prod(input_shape)).reshape(input_shape)
    >>> print(x)
    [[[ 0  1  2]
      [ 3  4  5]]
     [[ 6  7  8]
      [ 9 10 11]]]
    >>> y = tf.keras.layers.UpSampling1D(size=2)(x)
    >>> print(y)
    tf.Tensor(
      [[[ 0  1  2]
        [ 0  1  2]
        [ 3  4  5]
        [ 3  4  5]]
       [[ 6  7  8]
        [ 6  7  8]
        [ 9 10 11]
        [ 9 10 11]]], shape=(2, 4, 3), dtype=int64)

    Args:
      size: Integer. Upsampling factor.

    Input shape:
      3D tensor with shape: `(batch_size, steps, features)`.

    Output shape:
      3D tensor with shape: `(batch_size, upsampled_steps, features)`.
    """

    def __init__(self, size=2, **kwargs):
        super().__init__(**kwargs)
        self.size = int(size)
        self.input_spec = InputSpec(ndim=3)

    def compute_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape).as_list()
        size = (
            self.size * input_shape[1] if input_shape[1] is not None else None
        )
        return tf.TensorShape([input_shape[0], size, input_shape[2]])

    def call(self, inputs):
        output = backend.repeat_elements(inputs, self.size, axis=1)
        return output

    def get_config(self):
        config = {"size": self.size}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



"""Image classification with EfficientNetV2 architecture.

Adapted from the EfficientNetV2 Keras Application.
"""
import math

from tensorflow import keras

from keras.integration_test.models.input_spec import InputSpec

IMG_SIZE = (96, 96)
NUM_CLASSES = 5


def get_data_spec(batch_size):
    return (
        InputSpec((batch_size,) + IMG_SIZE + (3,)),
        InputSpec((batch_size, NUM_CLASSES)),
    )


def get_input_preprocessor():
    return keras.layers.Rescaling(scale=1.0 / 128.0, offset=-1)


def round_filters(filters, width_coefficient, min_depth, depth_divisor):
    filters *= width_coefficient
    minimum_depth = min_depth or depth_divisor
    new_filters = max(
        minimum_depth,
        int(filters + depth_divisor / 2) // depth_divisor * depth_divisor,
    )
    return int(new_filters)


def MBConvBlock(
    input_filters: int,
    output_filters: int,
    expand_ratio=1,
    kernel_size=3,
    strides=1,
    se_ratio=0.0,
    activation="swish",
    survival_probability: float = 0.8,
):
    def apply(inputs):
        filters = input_filters * expand_ratio
        if expand_ratio != 1:
            x = keras.layers.Conv2D(
                filters=filters,
                kernel_size=1,
                strides=1,
                padding="same",
                data_format="channels_last",
                use_bias=False,
            )(inputs)
            x = keras.layers.BatchNormalization()(x)
            x = keras.layers.Activation(activation)(x)
        else:
            x = inputs

        x = keras.layers.DepthwiseConv2D(
            kernel_size=kernel_size,
            strides=strides,
            padding="same",
            data_format="channels_last",
            use_bias=False,
        )(x)
        x = keras.layers.BatchNormalization()(x)
        x = keras.layers.Activation(activation)(x)

        if 0 < se_ratio <= 1:
            filters_se = max(1, int(input_filters * se_ratio))
            se = keras.layers.GlobalAveragePooling2D()(x)
            se = keras.layers.Reshape((1, 1, filters))(se)
            se = keras.layers.Conv2D(
                filters_se,
                1,
                padding="same",
                activation=activation,
            )(se)
            se = keras.layers.Conv2D(
                filters,
                1,
                padding="same",
                activation="sigmoid",
            )(se)
            x = keras.layers.multiply([x, se])
            x = keras.layers.Conv2D(
                filters=output_filters,
                kernel_size=1,
                strides=1,
                padding="same",
                data_format="channels_last",
                use_bias=False,
            )(x)
            x = keras.layers.BatchNormalization()(x)

            if strides == 1 and input_filters == output_filters:
                if survival_probability:
                    x = keras.layers.Dropout(
                        survival_probability,
                        noise_shape=(None, 1, 1, 1),
                    )(x)
                x = keras.layers.add([x, inputs])
        return x

    return apply


def FusedMBConvBlock(
    input_filters: int,
    output_filters: int,
    expand_ratio=1,
    kernel_size=3,
    strides=1,
    se_ratio=0.0,
    activation="swish",
    survival_probability: float = 0.8,
):
    def apply(inputs):
        filters = input_filters * expand_ratio
        if expand_ratio != 1:
            x = keras.layers.Conv2D(
                filters,
                kernel_size=kernel_size,
                strides=strides,
                data_format="channels_last",
                padding="same",
                use_bias=False,
            )(inputs)
            x = keras.layers.BatchNormalization()(x)
            x = keras.layers.Activation(activation)(x)
        else:
            x = inputs

        if 0 < se_ratio <= 1:
            filters_se = max(1, int(input_filters * se_ratio))
            se = keras.layers.GlobalAveragePooling2D()(x)
            se = keras.layers.Reshape((1, 1, filters))(se)
            se = keras.layers.Conv2D(
                filters_se,
                1,
                padding="same",
                activation=activation,
            )(se)
            se = keras.layers.Conv2D(
                filters,
                1,
                padding="same",
                activation="sigmoid",
            )(se)
            x = keras.layers.multiply([x, se])

        x = keras.layers.Conv2D(
            output_filters,
            kernel_size=1 if expand_ratio != 1 else kernel_size,
            strides=1 if expand_ratio != 1 else strides,
            padding="same",
            use_bias=False,
        )(x)
        x = keras.layers.BatchNormalization()(x)

        if expand_ratio == 1:
            x = keras.layers.Activation(activation)(x)

        if strides == 1 and input_filters == output_filters:
            if survival_probability:
                x = keras.layers.Dropout(
                    survival_probability,
                    noise_shape=(None, 1, 1, 1),
                )(x)
            x = keras.layers.add([x, inputs])

        return x

    return apply


def get_model(
    build=False, compile=False, jit_compile=False, include_preprocessing=True
):
    width_coefficient = 1.0
    depth_coefficient = 1.0
    dropout_rate = 0.2
    drop_connect_rate = 0.2
    depth_divisor = 8
    min_depth = 8
    activation = "swish"
    blocks_args = [
        {
            "kernel_size": 3,
            "num_repeat": 2,
            "input_filters": 24,
            "output_filters": 24,
            "expand_ratio": 1,
            "se_ratio": 0.0,
            "strides": 1,
            "conv_type": 1,
        },
        {
            "kernel_size": 3,
            "num_repeat": 4,
            "input_filters": 24,
            "output_filters": 48,
            "expand_ratio": 4,
            "se_ratio": 0.0,
            "strides": 2,
            "conv_type": 1,
        },
        {
            "conv_type": 1,
            "expand_ratio": 4,
            "input_filters": 48,
            "kernel_size": 3,
            "num_repeat": 4,
            "output_filters": 64,
            "se_ratio": 0,
            "strides": 2,
        },
        {
            "conv_type": 0,
            "expand_ratio": 4,
            "input_filters": 64,
            "kernel_size": 3,
            "num_repeat": 6,
            "output_filters": 128,
            "se_ratio": 0.25,
            "strides": 2,
        },
    ]

    inputs = keras.layers.Input(shape=IMG_SIZE + (3,))
    if include_preprocessing:
        x = get_input_preprocessor()(inputs)
    else:
        x = inputs

    stem_filters = round_filters(
        filters=blocks_args[0]["input_filters"],
        width_coefficient=width_coefficient,
        min_depth=min_depth,
        depth_divisor=depth_divisor,
    )
    x = keras.layers.Conv2D(
        filters=stem_filters,
        kernel_size=3,
        strides=2,
        padding="same",
        use_bias=False,
    )(x)
    x = keras.layers.BatchNormalization()(x)
    x = keras.layers.Activation(activation, name="stem_activation")(x)

    b = 0
    blocks = float(sum(args["num_repeat"] for args in blocks_args))
    for _, args in enumerate(blocks_args):
        args["input_filters"] = round_filters(
            filters=args["input_filters"],
            width_coefficient=width_coefficient,
            min_depth=min_depth,
            depth_divisor=depth_divisor,
        )
        args["output_filters"] = round_filters(
            filters=args["output_filters"],
            width_coefficient=width_coefficient,
            min_depth=min_depth,
            depth_divisor=depth_divisor,
        )
        block = {0: MBConvBlock, 1: FusedMBConvBlock}[args.pop("conv_type")]
        repeats = int(math.ceil(depth_coefficient * args.pop("num_repeat")))
        for j in range(repeats):
            if j > 0:
                args["strides"] = 1
                args["input_filters"] = args["output_filters"]

            x = block(
                activation=activation,
                survival_probability=drop_connect_rate * b / blocks,
                **args,
            )(x)
            b += 1

    top_filters = round_filters(
        filters=1280,
        width_coefficient=width_coefficient,
        min_depth=min_depth,
        depth_divisor=depth_divisor,
    )
    x = keras.layers.Conv2D(
        filters=top_filters,
        kernel_size=1,
        strides=1,
        padding="same",
        data_format="channels_last",
        use_bias=False,
    )(x)
    x = keras.layers.BatchNormalization()(x)
    x = keras.layers.Activation(activation=activation, name="top_activation")(x)
    x = keras.layers.GlobalAveragePooling2D(name="avg_pool")(x)
    x = keras.layers.Dropout(dropout_rate, name="top_dropout")(x)
    x = keras.layers.Dense(
        NUM_CLASSES,
        activation="softmax",
    )(x)
    model = keras.Model(inputs, x)
    if compile:
        model.compile(
            "adam", loss="categorical_crossentropy", jit_compile=jit_compile
        )
    return model


def get_custom_objects():
    return {}



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Classes and functions implementing to Network SavedModel serialization."""

from keras.saving.legacy.saved_model import constants
from keras.saving.legacy.saved_model import model_serialization


# FunctionalModel serialization is pretty much the same as Model serialization.
class NetworkSavedModelSaver(model_serialization.ModelSavedModelSaver):
    """Network serialization."""

    @property
    def object_identifier(self):
        return constants.NETWORK_IDENTIFIER



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Private utilities for locally-connected layers."""

import numpy as np
import tensorflow.compat.v2 as tf

from keras import backend
from keras.utils import conv_utils


def get_locallyconnected_mask(
    input_shape, kernel_shape, strides, padding, data_format
):
    """Return a mask representing connectivity of a locally-connected operation.

    This method returns a masking numpy array of 0s and 1s (of type
    `np.float32`) that, when element-wise multiplied with a fully-connected
    weight tensor, masks out the weights between disconnected input-output pairs
    and thus implements local connectivity through a sparse fully-connected
    weight tensor.

    Assume an unshared convolution with given parameters is applied to an input
    having N spatial dimensions with `input_shape = (d_in1, ..., d_inN)`
    to produce an output with spatial shape `(d_out1, ..., d_outN)` (determined
    by layer parameters such as `strides`).

    This method returns a mask which can be broadcast-multiplied (element-wise)
    with a 2*(N+1)-D weight matrix (equivalent to a fully-connected layer
    between (N+1)-D activations (N spatial + 1 channel dimensions for input and
    output) to make it perform an unshared convolution with given
    `kernel_shape`, `strides`, `padding` and `data_format`.

    Args:
      input_shape: tuple of size N: `(d_in1, ..., d_inN)` spatial shape of the
        input.
      kernel_shape: tuple of size N, spatial shape of the convolutional kernel /
        receptive field.
      strides: tuple of size N, strides along each spatial dimension.
      padding: type of padding, string `"same"` or `"valid"`.
      data_format: a string, `"channels_first"` or `"channels_last"`.

    Returns:
      a `np.float32`-type `np.ndarray` of shape
      `(1, d_in1, ..., d_inN, 1, d_out1, ..., d_outN)`
      if `data_format == `"channels_first"`, or
      `(d_in1, ..., d_inN, 1, d_out1, ..., d_outN, 1)`
      if `data_format == "channels_last"`.

    Raises:
      ValueError: if `data_format` is neither `"channels_first"` nor
                  `"channels_last"`.
    """
    mask = conv_utils.conv_kernel_mask(
        input_shape=input_shape,
        kernel_shape=kernel_shape,
        strides=strides,
        padding=padding,
    )

    ndims = int(mask.ndim / 2)

    if data_format == "channels_first":
        mask = np.expand_dims(mask, 0)
        mask = np.expand_dims(mask, -ndims - 1)

    elif data_format == "channels_last":
        mask = np.expand_dims(mask, ndims)
        mask = np.expand_dims(mask, -1)

    else:
        raise ValueError("Unrecognized data_format: " + str(data_format))

    return mask


def local_conv_matmul(inputs, kernel, kernel_mask, output_shape):
    """Apply N-D convolution with un-shared weights using a single matmul call.

    This method outputs `inputs . (kernel * kernel_mask)`
    (with `.` standing for matrix-multiply and `*` for element-wise multiply)
    and requires a precomputed `kernel_mask` to zero-out weights in `kernel` and
    hence perform the same operation as a convolution with un-shared
    (the remaining entries in `kernel`) weights. It also does the necessary
    reshapes to make `inputs` and `kernel` 2-D and `output` (N+2)-D.

    Args:
        inputs: (N+2)-D tensor with shape `(batch_size, channels_in, d_in1, ...,
          d_inN)` or `(batch_size, d_in1, ..., d_inN, channels_in)`.
        kernel: the unshared weights for N-D convolution,
            an (N+2)-D tensor of shape: `(d_in1, ..., d_inN, channels_in,
            d_out2, ..., d_outN, channels_out)` or `(channels_in, d_in1, ...,
            d_inN, channels_out, d_out2, ..., d_outN)`, with the ordering of
            channels and spatial dimensions matching that of the input. Each
            entry is the weight between a particular input and output location,
            similarly to a fully-connected weight matrix.
        kernel_mask: a float 0/1 mask tensor of shape: `(d_in1, ..., d_inN, 1,
          d_out2, ..., d_outN, 1)` or `(1, d_in1, ..., d_inN, 1, d_out2, ...,
          d_outN)`, with the ordering of singleton and spatial dimensions
          matching that of the input. Mask represents the connectivity pattern
          of the layer and is precomputed elsewhere based on layer parameters:
          stride, padding, and the receptive field shape.
        output_shape: a tuple of (N+2) elements representing the output shape:
          `(batch_size, channels_out, d_out1, ..., d_outN)` or `(batch_size,
          d_out1, ..., d_outN, channels_out)`, with the ordering of channels and
          spatial dimensions matching that of the input.

    Returns:
        Output (N+2)-D tensor with shape `output_shape`.
    """
    inputs_flat = backend.reshape(inputs, (backend.shape(inputs)[0], -1))

    kernel = kernel_mask * kernel
    kernel = make_2d(kernel, split_dim=backend.ndim(kernel) // 2)

    output_flat = tf.matmul(inputs_flat, kernel, b_is_sparse=True)
    output = backend.reshape(
        output_flat,
        [
            backend.shape(output_flat)[0],
        ]
        + output_shape.as_list()[1:],
    )
    return output


def local_conv_sparse_matmul(
    inputs, kernel, kernel_idxs, kernel_shape, output_shape
):
    """Apply N-D convolution with unshared weights using a single sparse matmul.

    This method outputs `inputs . tf.sparse.SparseTensor(indices=kernel_idxs,
    values=kernel, dense_shape=kernel_shape)`, with `.` standing for
    matrix-multiply. It also reshapes `inputs` to 2-D and `output` to (N+2)-D.

    Args:
        inputs: (N+2)-D tensor with shape `(batch_size, channels_in, d_in1, ...,
          d_inN)` or `(batch_size, d_in1, ..., d_inN, channels_in)`.
        kernel: a 1-D tensor with shape `(len(kernel_idxs),)` containing all the
          weights of the layer.
        kernel_idxs:  a list of integer tuples representing indices in a sparse
          matrix performing the un-shared convolution as a matrix-multiply.
        kernel_shape: a tuple `(input_size, output_size)`, where `input_size =
          channels_in * d_in1 * ... * d_inN` and `output_size = channels_out *
          d_out1 * ... * d_outN`.
        output_shape: a tuple of (N+2) elements representing the output shape:
          `(batch_size, channels_out, d_out1, ..., d_outN)` or `(batch_size,
          d_out1, ..., d_outN, channels_out)`, with the ordering of channels and
          spatial dimensions matching that of the input.

    Returns:
        Output (N+2)-D dense tensor with shape `output_shape`.
    """
    inputs_flat = backend.reshape(inputs, (backend.shape(inputs)[0], -1))
    output_flat = tf.sparse.sparse_dense_matmul(
        sp_a=tf.SparseTensor(kernel_idxs, kernel, kernel_shape),
        b=inputs_flat,
        adjoint_b=True,
    )
    output_flat_transpose = backend.transpose(output_flat)

    output_reshaped = backend.reshape(
        output_flat_transpose,
        [
            backend.shape(output_flat_transpose)[0],
        ]
        + output_shape.as_list()[1:],
    )
    return output_reshaped


def make_2d(tensor, split_dim):
    """Reshapes an N-dimensional tensor into a 2D tensor.

    Dimensions before (excluding) and after (including) `split_dim` are grouped
    together.

    Args:
      tensor: a tensor of shape `(d0, ..., d(N-1))`.
      split_dim: an integer from 1 to N-1, index of the dimension to group
        dimensions before (excluding) and after (including).

    Returns:
      Tensor of shape
      `(d0 * ... * d(split_dim-1), d(split_dim) * ... * d(N-1))`.
    """
    shape = tf.shape(tensor)
    in_dims = shape[:split_dim]
    out_dims = shape[split_dim:]

    in_size = tf.reduce_prod(in_dims)
    out_size = tf.reduce_prod(out_dims)

    return tf.reshape(tensor, (in_size, out_size))



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Correctness tests for tf.keras CNN models using DistributionStrategy."""

import numpy as np
import tensorflow.compat.v2 as tf

import keras
from keras.distribute import keras_correctness_test_base
from keras.optimizers.legacy import gradient_descent
from keras.testing_infra import test_utils


@test_utils.run_all_without_tensor_float_32(
    "Uses Dense layers, which call matmul. Even if Dense layers run in "
    "float64, the test sometimes fails with TensorFloat-32 enabled for unknown "
    "reasons"
)
@test_utils.run_v2_only()
class DistributionStrategyCnnCorrectnessTest(
    keras_correctness_test_base.TestDistributionStrategyCorrectnessBase
):
    def get_model(
        self, initial_weights=None, distribution=None, input_shapes=None
    ):
        del input_shapes
        with keras_correctness_test_base.MaybeDistributionScope(distribution):
            image = keras.layers.Input(shape=(28, 28, 3), name="image")
            c1 = keras.layers.Conv2D(
                name="conv1",
                filters=16,
                kernel_size=(3, 3),
                strides=(4, 4),
                kernel_regularizer=keras.regularizers.l2(1e-4),
            )(image)
            if self.with_batch_norm == "regular":
                c1 = keras.layers.BatchNormalization(name="bn1")(c1)
            elif self.with_batch_norm == "sync":
                # Test with parallel batch norms to verify all-reduce works OK.
                bn1 = keras.layers.BatchNormalization(
                    name="bn1", synchronized=True
                )(c1)
                bn2 = keras.layers.BatchNormalization(
                    name="bn2", synchronized=True
                )(c1)
                c1 = keras.layers.Add()([bn1, bn2])
            c1 = keras.layers.MaxPooling2D(pool_size=(2, 2))(c1)
            logits = keras.layers.Dense(10, activation="softmax", name="pred")(
                keras.layers.Flatten()(c1)
            )
            model = keras.Model(inputs=[image], outputs=[logits])

            if initial_weights:
                model.set_weights(initial_weights)

            model.compile(
                optimizer=gradient_descent.SGD(learning_rate=0.1),
                loss="sparse_categorical_crossentropy",
                metrics=["sparse_categorical_accuracy"],
            )

        return model

    def _get_data(self, count, shape=(28, 28, 3), num_classes=10):
        centers = np.random.randn(num_classes, *shape)

        features = []
        labels = []
        for _ in range(count):
            label = np.random.randint(0, num_classes, size=1)[0]
            offset = np.random.normal(loc=0, scale=0.1, size=np.prod(shape))
            offset = offset.reshape(shape)
            labels.append(label)
            features.append(centers[label] + offset)

        x = np.asarray(features, dtype=np.float32)
        y = np.asarray(labels, dtype=np.float32).reshape((count, 1))
        return x, y

    def get_data(self):
        x_train, y_train = self._get_data(
            count=keras_correctness_test_base._GLOBAL_BATCH_SIZE
            * keras_correctness_test_base._EVAL_STEPS
        )
        x_predict = x_train
        return x_train, y_train, x_predict

    def get_data_with_partial_last_batch_eval(self):
        x_train, y_train = self._get_data(count=1280)
        x_eval, y_eval = self._get_data(count=1000)
        return x_train, y_train, x_eval, y_eval, x_eval

    @tf.__internal__.distribute.combinations.generate(
        keras_correctness_test_base.all_strategy_and_input_config_combinations()
        + keras_correctness_test_base.multi_worker_mirrored_eager()
    )
    def test_cnn_correctness(
        self, distribution, use_numpy, use_validation_data
    ):
        if (
            distribution
            == tf.__internal__.distribute.combinations.central_storage_strategy_with_gpu_and_cpu  # noqa: E501
        ):
            self.skipTest("b/183958183")
        self.run_correctness_test(distribution, use_numpy, use_validation_data)

    @tf.__internal__.distribute.combinations.generate(
        keras_correctness_test_base.all_strategy_and_input_config_combinations()
        + keras_correctness_test_base.multi_worker_mirrored_eager()
    )
    def test_cnn_with_batch_norm_correctness(
        self, distribution, use_numpy, use_validation_data
    ):
        self.run_correctness_test(
            distribution,
            use_numpy,
            use_validation_data,
            with_batch_norm="regular",
        )

    @tf.__internal__.distribute.combinations.generate(
        keras_correctness_test_base.all_strategy_and_input_config_combinations()
        + keras_correctness_test_base.multi_worker_mirrored_eager()
    )
    def test_cnn_with_sync_batch_norm_correctness(
        self, distribution, use_numpy, use_validation_data
    ):
        if not tf.executing_eagerly():
            self.skipTest(
                "BatchNorm with `synchronized` is not enabled in graph mode."
            )
        self.run_correctness_test(
            distribution, use_numpy, use_validation_data, with_batch_norm="sync"
        )

    @tf.__internal__.distribute.combinations.generate(
        keras_correctness_test_base.all_strategy_and_input_config_combinations_eager()  # noqa: E501
        + keras_correctness_test_base.multi_worker_mirrored_eager()
        + keras_correctness_test_base.test_combinations_with_tpu_strategies_graph()  # noqa: E501
    )
    def test_cnn_correctness_with_partial_last_batch_eval(
        self, distribution, use_numpy, use_validation_data
    ):
        self.run_correctness_test(
            distribution,
            use_numpy,
            use_validation_data,
            partial_last_batch=True,
            training_epochs=1,
        )

    @tf.__internal__.distribute.combinations.generate(
        keras_correctness_test_base.all_strategy_and_input_config_combinations_eager()  # noqa: E501
        + keras_correctness_test_base.multi_worker_mirrored_eager()
        + keras_correctness_test_base.test_combinations_with_tpu_strategies_graph()  # noqa: E501
    )
    def test_cnn_with_batch_norm_correctness_and_partial_last_batch_eval(
        self, distribution, use_numpy, use_validation_data
    ):
        self.run_correctness_test(
            distribution,
            use_numpy,
            use_validation_data,
            with_batch_norm="regular",
            partial_last_batch=True,
        )


if __name__ == "__main__":
    tf.__internal__.distribute.multi_process_runner.test_main()



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Keras utilities for DTensor unit test."""

import numpy as np
import tensorflow.compat.v2 as tf
from absl.testing import parameterized

# isort: off
from tensorflow.dtensor.python import api as dtensor_api
from tensorflow.python.eager import context

_DEFAULT_GPU_MEMORY_LIMIT = 200  # MB


class DTensorBaseTest(tf.test.TestCase, parameterized.TestCase):
    """Provides comparison helper for dtensor vs local results."""

    @classmethod
    def setUpClass(cls):
        super(DTensorBaseTest, cls).setUpClass()

    def tearDown(self):
        super().tearDown()
        # Make sure all async ops finish.
        context.async_wait()

        # TODO(hthu): Remove the reset once we fixed the CopyToMesh with
        # DefaultMesh placement issue.
        reset_dtensor()

    @staticmethod
    def configTestMesh(device_type_mesh_map):
        """Configs corresponding mesh given test context.

        If runs on a CPU mesh, set virtual device on CPU.
        If runs on a GPU mesh, sets virtual device on GPU with proper memory
        limits.
        if runs on a TPU mesh, initializes TPU system.

        Args:
          device_type_mesh_map: A dictionary containing device_type -> mesh
            mapping.

        Returns:
          A properly configured mesh for use in test.
        """
        reset_context()

        def get_mesh(device_type):
            mesh = device_type_mesh_map.get(device_type, None)
            if mesh is None:
                dt = device_type
                raise ValueError(f"Requires a {dt} mesh to run test on {dt}.")
            return mesh

        mesh = None
        if tf.config.list_physical_devices("GPU"):
            mesh = get_mesh("GPU")
            reset_logical_devices("GPU", np.prod(mesh.shape()))
        else:
            mesh = get_mesh("CPU")
            reset_logical_devices("CPU", np.prod(mesh.shape()))

        context.ensure_initialized()
        return mesh


def create_device_array(shape, device_type):
    device_count = np.prod(shape)
    return np.asarray(
        [
            tf.DeviceSpec(
                job="localhost/replica:0/task:0",
                device_type=device_type,
                device_index=i,
            )
            for i in range(device_count)
        ]
    ).reshape(shape)


def create_device_list(shape, device_type):
    devices = create_device_array(shape, device_type)
    return np.ravel(devices).tolist()


def create_device_ids_array(shape):
    device_count = np.prod(shape)
    return np.arange(device_count).reshape(shape)


def reset_context():
    context._reset_context()


def reset_logical_devices(device_type, count):
    """Resets logical devices for CPU/GPU.

    Logical devices can only be instantiated once on a particular context. For
    now, context re-use is triggering some function duplication errors, so we
    reset the context on each call.

    Args:
      device_type: The device_type to reset.
      count: numbers of virtual device to reset to.
    """
    reset_context()
    devices = tf.config.list_physical_devices(device_type)
    if device_type.upper() == "CPU":
        tf.config.set_logical_device_configuration(
            devices[0],
            [
                tf.config.LogicalDeviceConfiguration(),
            ]
            * count,
        )
    elif device_type.upper() == "GPU":
        tf.config.set_logical_device_configuration(
            devices[0],
            [
                tf.config.LogicalDeviceConfiguration(
                    memory_limit=_DEFAULT_GPU_MEMORY_LIMIT
                ),
            ]
            * count,
        )
    else:
        dt = device_type
        raise ValueError(
            f"resetting logical device for non-supported device type: {dt}"
        )


def reset_dtensor():
    dtensor_api._reset()



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Wrapper layer to apply every temporal slice of an input."""


import tensorflow.compat.v2 as tf

from keras import backend
from keras.engine.base_layer import Layer
from keras.engine.input_spec import InputSpec
from keras.layers.rnn.base_wrapper import Wrapper
from keras.utils import generic_utils
from keras.utils import layer_utils
from keras.utils import tf_utils

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.layers.TimeDistributed")
class TimeDistributed(Wrapper):
    """This wrapper allows to apply a layer to every temporal slice of an input.

    Every input should be at least 3D, and the dimension of index one of the
    first input will be considered to be the temporal dimension.

    Consider a batch of 32 video samples, where each sample is a 128x128 RGB
    image with `channels_last` data format, across 10 timesteps.
    The batch input shape is `(32, 10, 128, 128, 3)`.

    You can then use `TimeDistributed` to apply the same `Conv2D` layer to each
    of the 10 timesteps, independently:

    >>> inputs = tf.keras.Input(shape=(10, 128, 128, 3))
    >>> conv_2d_layer = tf.keras.layers.Conv2D(64, (3, 3))
    >>> outputs = tf.keras.layers.TimeDistributed(conv_2d_layer)(inputs)
    >>> outputs.shape
    TensorShape([None, 10, 126, 126, 64])

    Because `TimeDistributed` applies the same instance of `Conv2D` to each of
    the timestamps, the same set of weights are used at each timestamp.

    Args:
      layer: a `tf.keras.layers.Layer` instance.

    Call arguments:
      inputs: Input tensor of shape (batch, time, ...) or nested tensors,
        and each of which has shape (batch, time, ...).
      training: Python boolean indicating whether the layer should behave in
        training mode or in inference mode. This argument is passed to the
        wrapped layer (only if the layer supports this argument).
      mask: Binary tensor of shape `(samples, timesteps)` indicating whether
        a given timestep should be masked. This argument is passed to the
        wrapped layer (only if the layer supports this argument).

    Raises:
      ValueError: If not initialized with a `tf.keras.layers.Layer` instance.
    """

    def __init__(self, layer, **kwargs):
        if not isinstance(layer, Layer):
            raise ValueError(
                "Please initialize `TimeDistributed` layer with a "
                f"`tf.keras.layers.Layer` instance. Received: {layer}"
            )
        super().__init__(layer, **kwargs)
        self.supports_masking = True

        # It is safe to use the fast, reshape-based approach with all of our
        # built-in Layers.
        self._always_use_reshape = layer_utils.is_builtin_layer(
            layer
        ) and not getattr(layer, "stateful", False)

    def _get_shape_tuple(self, init_tuple, tensor, start_idx):
        """Finds non-specific dimensions in the static shapes.

        The static shapes are replaced with the corresponding dynamic shapes of
        the tensor.
        Args:
          init_tuple: a tuple, the first part of the output shape
          tensor: the tensor from which to get the (static and dynamic) shapes
            as the last part of the output shape
          start_idx: int, which indicate the first dimension to take from
            the static shape of the tensor
        Returns:
          The new shape with the first part from `init_tuple` and the last part
          from or `tensor.shape`, where every `None` is replaced by the
          corresponding dimension from `tf.shape(tensor)`.
        """
        # replace all None in int_shape by backend.shape
        int_shape = backend.int_shape(tensor)[start_idx:]
        if not any(s is None for s in int_shape):
            return init_tuple + int_shape
        shape = backend.shape(tensor)
        int_shape = list(int_shape)
        for i, s in enumerate(int_shape):
            if s is None:
                int_shape[i] = shape[start_idx + i]
        return init_tuple + tuple(int_shape)

    def _remove_timesteps(self, dims):
        dims = dims.as_list()
        return tf.TensorShape([dims[0]] + dims[2:])

    def build(self, input_shape):
        input_shape = tf_utils.convert_shapes(input_shape, to_tuples=False)
        input_dims = tf.nest.flatten(
            tf.nest.map_structure(lambda x: x.ndims, input_shape)
        )
        if any(dim < 3 for dim in input_dims):
            raise ValueError(
                "`TimeDistributed` Layer should be passed an `input_shape ` "
                f"with at least 3 dimensions, received: {input_shape}"
            )
        # Don't enforce the batch or time dimension.
        self.input_spec = tf.nest.map_structure(
            lambda x: InputSpec(shape=[None, None] + x.as_list()[2:]),
            input_shape,
        )
        child_input_shape = tf.nest.map_structure(
            self._remove_timesteps, input_shape
        )
        child_input_shape = tf_utils.convert_shapes(child_input_shape)
        super().build(tuple(child_input_shape))
        self.built = True

    def compute_output_shape(self, input_shape):
        input_shape = tf_utils.convert_shapes(input_shape, to_tuples=False)

        child_input_shape = tf.nest.map_structure(
            self._remove_timesteps, input_shape
        )
        child_output_shape = self.layer.compute_output_shape(child_input_shape)
        child_output_shape = tf_utils.convert_shapes(
            child_output_shape, to_tuples=False
        )
        timesteps = tf_utils.convert_shapes(input_shape)
        timesteps = tf.nest.flatten(timesteps)[1]

        def insert_timesteps(dims):
            dims = dims.as_list()
            return tf.TensorShape([dims[0], timesteps] + dims[1:])

        return tf.nest.map_structure(insert_timesteps, child_output_shape)

    def call(self, inputs, training=None, mask=None):
        kwargs = {}
        if generic_utils.has_arg(self.layer.call, "training"):
            kwargs["training"] = training

        input_shape = tf.nest.map_structure(
            lambda x: tf.TensorShape(backend.int_shape(x)), inputs
        )
        batch_size = tf_utils.convert_shapes(input_shape)
        batch_size = tf.nest.flatten(batch_size)[0]
        if batch_size and not self._always_use_reshape:
            inputs, row_lengths = backend.convert_inputs_if_ragged(inputs)
            is_ragged_input = row_lengths is not None
            input_length = tf_utils.convert_shapes(input_shape)
            input_length = tf.nest.flatten(input_length)[1]

            # batch size matters, use rnn-based implementation
            def step(x, _):
                output = self.layer(x, **kwargs)
                return output, []

            _, outputs, _ = backend.rnn(
                step,
                inputs,
                initial_states=[],
                input_length=row_lengths[0]
                if is_ragged_input
                else input_length,
                mask=mask,
                unroll=False,
            )

            y = tf.nest.map_structure(
                lambda output: backend.maybe_convert_to_ragged(
                    is_ragged_input, output, row_lengths
                ),
                outputs,
            )
        else:
            # No batch size specified, therefore the layer will be able
            # to process batches of any size.
            # We can go with reshape-based implementation for performance.
            is_ragged_input = tf.nest.map_structure(
                lambda x: isinstance(x, tf.RaggedTensor), inputs
            )
            is_ragged_input = tf.nest.flatten(is_ragged_input)
            if all(is_ragged_input):
                input_values = tf.nest.map_structure(lambda x: x.values, inputs)
                input_row_lenghts = tf.nest.map_structure(
                    lambda x: x.nested_row_lengths()[0], inputs
                )
                y = self.layer(input_values, **kwargs)
                y = tf.nest.map_structure(
                    tf.RaggedTensor.from_row_lengths, y, input_row_lenghts
                )
            elif any(is_ragged_input):
                raise ValueError(
                    "All inputs has to be either ragged or not, "
                    f"but not mixed. Received: {inputs}"
                )
            else:
                input_length = tf_utils.convert_shapes(input_shape)
                input_length = tf.nest.flatten(input_length)[1]
                if not input_length:
                    input_length = tf.nest.map_structure(
                        lambda x: tf.shape(x)[1], inputs
                    )
                    input_length = generic_utils.to_list(
                        tf.nest.flatten(input_length)
                    )[0]

                inner_input_shape = tf.nest.map_structure(
                    lambda x: self._get_shape_tuple((-1,), x, 2), inputs
                )
                # Shape: (num_samples * timesteps, ...). And track the
                # transformation in self._input_map.
                inputs = tf.__internal__.nest.map_structure_up_to(
                    inputs, tf.reshape, inputs, inner_input_shape
                )
                # (num_samples * timesteps, ...)
                if (
                    generic_utils.has_arg(self.layer.call, "mask")
                    and mask is not None
                ):
                    inner_mask_shape = self._get_shape_tuple((-1,), mask, 2)
                    kwargs["mask"] = backend.reshape(mask, inner_mask_shape)

                y = self.layer(inputs, **kwargs)

                # Reconstruct the output shape by re-splitting the 0th dimension
                # back into (num_samples, timesteps, ...)
                # We use batch_size when available so that the 0th dimension is
                # set in the static shape of the reshaped output
                reshape_batch_size = batch_size if batch_size else -1
                output_shape = tf.nest.map_structure(
                    lambda tensor: self._get_shape_tuple(
                        (reshape_batch_size, input_length), tensor, 1
                    ),
                    y,
                )
                y = tf.__internal__.nest.map_structure_up_to(
                    y, tf.reshape, y, output_shape
                )

        return y

    def compute_mask(self, inputs, mask=None):
        """Computes an output mask tensor for Embedding layer.

        This is based on the inputs, mask, and the inner layer.
        If batch size is specified:
        Simply return the input `mask`. (An rnn-based implementation with
        more than one rnn inputs is required but not supported in tf.keras yet.)
        Otherwise we call `compute_mask` of the inner layer at each time step.
        If the output mask at each time step is not `None`:
        (E.g., inner layer is Masking or RNN)
        Concatenate all of them and return the concatenation.
        If the output mask at each time step is `None` and the input mask is not
        `None`:(E.g., inner layer is Dense)
        Reduce the input_mask to 2 dimensions and return it.
        Otherwise (both the output mask and the input mask are `None`):
        (E.g., `mask` is not used at all)
        Return `None`.

        Args:
          inputs: Tensor with shape [batch size, timesteps, ...] indicating the
            input to TimeDistributed. If static shape information is available
            for "batch size", `mask` is returned unmodified.
          mask: Either None (indicating no masking) or a Tensor indicating the
            input mask for TimeDistributed. The shape can be static or dynamic.

        Returns:
          Either None (no masking), or a [batch size, timesteps, ...] Tensor
          with an output mask for the TimeDistributed layer with the shape
          beyond the second dimension being the value of the input mask shape(if
          the computed output mask is none), an output mask with the shape
          beyond the first dimension being the value of the mask shape(if mask
          is not None) or output mask with the shape beyond the first dimension
          being the value of the computed output shape.

        """
        # cases need to call the layer.compute_mask when input_mask is None:
        # Masking layer and Embedding layer with mask_zero
        input_shape = tf.nest.map_structure(
            lambda x: tf.TensorShape(backend.int_shape(x)), inputs
        )
        input_shape = tf_utils.convert_shapes(input_shape, to_tuples=False)
        batch_size = tf_utils.convert_shapes(input_shape)
        batch_size = tf.nest.flatten(batch_size)[0]
        is_ragged_input = tf.nest.map_structure(
            lambda x: isinstance(x, tf.RaggedTensor), inputs
        )
        is_ragged_input = generic_utils.to_list(
            tf.nest.flatten(is_ragged_input)
        )
        if batch_size and not self._always_use_reshape or any(is_ragged_input):
            # batch size matters, we currently do not handle mask explicitly, or
            # if the layer always uses reshape approach, or the input is a
            # ragged tensor.
            return mask
        inner_mask = mask
        if inner_mask is not None:
            inner_mask_shape = self._get_shape_tuple((-1,), mask, 2)
            inner_mask = backend.reshape(inner_mask, inner_mask_shape)
        inner_input_shape = tf.nest.map_structure(
            lambda tensor: self._get_shape_tuple((-1,), tensor, 2), inputs
        )
        inner_inputs = tf.__internal__.nest.map_structure_up_to(
            inputs, tf.reshape, inputs, inner_input_shape
        )
        output_mask = self.layer.compute_mask(inner_inputs, inner_mask)
        if output_mask is None:
            if mask is None:
                return None
            # input_mask is not None, and output_mask is None:
            # we should return a not-None mask
            output_mask = mask
            for _ in range(2, len(backend.int_shape(mask))):
                output_mask = backend.any(output_mask, axis=-1)
        else:
            # output_mask is not None. We need to reshape it
            input_length = tf_utils.convert_shapes(input_shape)
            input_length = tf.nest.flatten(input_length)[1]
            if not input_length:
                input_length = tf.nest.map_structure(
                    lambda x: backend.shape(x)[1], inputs
                )
                input_length = tf.nest.flatten(input_length)[0]
            reshape_batch_size = batch_size if batch_size else -1
            output_mask_shape = self._get_shape_tuple(
                (reshape_batch_size, input_length), output_mask, 1
            )
            output_mask = backend.reshape(output_mask, output_mask_shape)
        return output_mask



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Hinge metrics."""

from keras.dtensor import utils as dtensor_utils
from keras.losses import categorical_hinge
from keras.losses import hinge
from keras.losses import squared_hinge
from keras.metrics import base_metric

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.metrics.Hinge")
class Hinge(base_metric.MeanMetricWrapper):
    """Computes the hinge metric between `y_true` and `y_pred`.

    `y_true` values are expected to be -1 or 1. If binary (0 or 1) labels are
    provided we will convert them to -1 or 1.

    Args:
      name: (Optional) string name of the metric instance.
      dtype: (Optional) data type of the metric result.

    Standalone usage:

    >>> m = tf.keras.metrics.Hinge()
    >>> m.update_state([[0, 1], [0, 0]], [[0.6, 0.4], [0.4, 0.6]])
    >>> m.result().numpy()
    1.3

    >>> m.reset_state()
    >>> m.update_state([[0, 1], [0, 0]], [[0.6, 0.4], [0.4, 0.6]],
    ...                sample_weight=[1, 0])
    >>> m.result().numpy()
    1.1

    Usage with `compile()` API:

    ```python
    model.compile(
        optimizer='sgd', loss='mse', metrics=[tf.keras.metrics.Hinge()])
    ```
    """

    @dtensor_utils.inject_mesh
    def __init__(self, name="hinge", dtype=None):
        super().__init__(hinge, name, dtype=dtype)


@keras_export("keras.metrics.SquaredHinge")
class SquaredHinge(base_metric.MeanMetricWrapper):
    """Computes the squared hinge metric between `y_true` and `y_pred`.

    `y_true` values are expected to be -1 or 1. If binary (0 or 1) labels are
    provided we will convert them to -1 or 1.

    Args:
      name: (Optional) string name of the metric instance.
      dtype: (Optional) data type of the metric result.

    Standalone usage:

    >>> m = tf.keras.metrics.SquaredHinge()
    >>> m.update_state([[0, 1], [0, 0]], [[0.6, 0.4], [0.4, 0.6]])
    >>> m.result().numpy()
    1.86

    >>> m.reset_state()
    >>> m.update_state([[0, 1], [0, 0]], [[0.6, 0.4], [0.4, 0.6]],
    ...                sample_weight=[1, 0])
    >>> m.result().numpy()
    1.46

    Usage with `compile()` API:

    ```python
    model.compile(
        optimizer='sgd',
        loss='mse',
        metrics=[tf.keras.metrics.SquaredHinge()])
    ```
    """

    @dtensor_utils.inject_mesh
    def __init__(self, name="squared_hinge", dtype=None):
        super().__init__(squared_hinge, name, dtype=dtype)


@keras_export("keras.metrics.CategoricalHinge")
class CategoricalHinge(base_metric.MeanMetricWrapper):
    """Computes the categorical hinge metric between `y_true` and `y_pred`.

    Args:
      name: (Optional) string name of the metric instance.
      dtype: (Optional) data type of the metric result.

    Standalone usage:

    >>> m = tf.keras.metrics.CategoricalHinge()
    >>> m.update_state([[0, 1], [0, 0]], [[0.6, 0.4], [0.4, 0.6]])
    >>> m.result().numpy()
    1.4000001

    >>> m.reset_state()
    >>> m.update_state([[0, 1], [0, 0]], [[0.6, 0.4], [0.4, 0.6]],
    ...                sample_weight=[1, 0])
    >>> m.result().numpy()
    1.2

    Usage with `compile()` API:

    ```python
    model.compile(
        optimizer='sgd',
        loss='mse',
        metrics=[tf.keras.metrics.CategoricalHinge()])
    ```
    """

    @dtensor_utils.inject_mesh
    def __init__(self, name="categorical_hinge", dtype=None):
        super().__init__(categorical_hinge, name, dtype=dtype)
